# Copyright 2016 Fred Hutchinson Cancer Research Center

import os

import pydot



class Digraph:
    def __init__(self):
        self.node_outs = { }
        self.node_ins = { }
    def has_node(self, node):
        return node in self.node_outs;
    def has_edge(self, edge):
        n1, n2 = edge
        return n1 in self.node_outs and n2 in self.node_outs[n1]
    def add_node(self, node):
        if (node not in self.node_outs):
              self.node_outs[node] = set();
              self.node_ins[node] = set();
    def add_nodes(self, nodes):
        for n in nodes:
              self.add_node(n);
    def add_edge(self, edge):
        n1,n2 = edge;
        self.node_outs[n1].add(n2);
        self.node_ins[n2].add(n1);
    def neighbors(self, node):
        # return list( self.nodes[node] );
        return self.node_outs[node];
    def __contains__(self, t):
        if (not self.has_node(t[0]) or not self.has_node(t[1])):
            return False;
        return t[1] in self.node_outs[t[0]];

    def nodes(self):
        return self.node_outs.keys();

    def find_sources(self):
        """return list of nodes that are source nodes, nothing points to them
        in the digraph, and they have nodes they point to"""
        nodes = [ n for n in self.nodes() if 0 == len(self.node_ins[n]) and 0 != len(self.node_outs[n]) ]
        return nodes;

    def find_sinks(self):
        """return list of nodes that are sink nodes, they don't point to any other node in the diagraph, and something points to them"""
        nodes = [ n for n in self.nodes() if 0 != len(self.node_ins[n]) and 0 == len(self.node_outs[n]) ]
        return nodes;






class TopologicalSort:
    def __init__(self, D):
        self.unmarked = set(D.nodes())
        self.D = D;
        self.temporary = set()
        self.marked = set();
        self.L = []

    def topological_sort(self):
        while len(self.unmarked) > 0:
            n = self.unmarked.pop();
            self.visit(n);

        self.L.reverse()
        return self.L

    def visit(self, node):
        if node in self.temporary:
            raise Exception("ERROR graph is not acyclic");
        if node in self.marked:
            return;
        self.temporary.add(node)
        for m in self.D.neighbors(node):
            self.visit(m)
        self.marked.add(node)
        self.temporary.remove(node);
        self.L.append(node);





def generate_processing_graph(processing_modules):
    D = Digraph()
    for m in processing_modules:
        D.add_node(m);
        for r in m.requires:
            D.add_node(r);
            D.add_edge( (r,m) );
        for p in m.provides:
            D.add_node(p);
            D.add_edge( (m,p) );
    return D;


def minimally_distinguishable_form(pathnames):
    def fullsplit(pathname):
        """Split path directories, and keep non-empty components"""
        if '' == pathname or '/' == pathname:
            return []
        head, tail = os.path.split(pathname)
        if '' == head:
            return [tail]
        elif '' != tail:
            return fullsplit(head) + [ tail]
        else:
            return fullsplit(head)

    def path_forms(path_list):
        """possible minimal forms of the path name using the list of components"""
        return [ os.path.join(*path_list[i:]) for i in range(len(path_list))]

    graph1 = {} # re-forumalating the full-path to path candidates as a dictionary
    for p in pathnames:
        for f in path_forms(fullsplit(p)):
            if f in graph1:
                graph1[f].append(p)
            else:
                graph1[f] = [p]
    groups = {}
    for candidate, names in graph1.items():
        if 1 != len(names):
            continue
        for name in names:
            if name in groups:
                groups[name].append(candidate)
            else:
                groups[name] = [candidate]
    mapping_table = {}
    for name, candidates in groups.items():
        mapping_table[name] = sorted(candidates, key=lambda x: len(x))[0]
    return mapping_table
            



def plot_graph(output_file, D):
    # graph = pydot.Dot(graph_type='digraph', nodesep=".75");
    graph = pydot.Dot(graph_type='digraph', ratio="compress");
    # graph.set_node_defaults(style="filled", fillcolor="grey")
    graph.set_edge_defaults(color="blue", arrowhead="vee", weight="1")

    # isinstance(x, str) ...

    path_names = [ n for n in D.nodes() if isinstance(n, str) ]
    minimal_table = minimally_distinguishable_form(path_names)


    node_to_pydot = {}
    for n in D.nodes():
        if isinstance(n, str):
            node_to_pydot[n] = pydot.Node( minimal_table[n], style='filled', fillcolor='lightblue', shape='box')
        else:
            # print n
            node_to_pydot[n] = pydot.Node( str(n), style='filled', fillcolor='grey')
        graph.add_node(node_to_pydot[n])


    for a in D.nodes():
        for b in D.neighbors(a):
            edge = pydot.Edge(node_to_pydot[a], node_to_pydot[b]);
            graph.add_edge(edge);
    graph.write_png(output_file);





"""YAGDSPY - Yet Another Graph Dependancy System in PYthon. Provide
core functions for building processing pipelines, whether for building
like make/maven/scons, or multi-step analysis using files to save
intermediate or final results.
"""
# Copyright 2016, 2017 Fred Hutchinson Cancer Research Center

# This file is to tell setuptools that this directory is a package

from .processing_module import *




# Copyright 2016, 2017 Fred Hutchinson Cancer Research Center
################################################################################
### Framework of processing module method


import os
import inspect


from .eval_processing_graph import *

class ProcessingModule:
    """Define basic idea of processing module class to be something that
    has a set of required files, and provides a set of files, while doing
    some kind of computation in the middle"""

    def __init__(self, base, config):
        self.base = base;
        self.config = config
        self.requires = set()
        self.provides = set()
        self.sub_modules = []
        self.name = "unknown, needs to be set"


    def __str__(self):
        return self.name;

    def is_readable(self, filename):
        return os.access(filename,  os.R_OK)

    def check_exist(self, fileset, required):
        timestamps = []
        exists = True
        for filename in fileset:
            if self.is_readable(filename):
                timestamps.append( os.path.getmtime(filename) )
            elif required:
                exists = False;
                raise Exception("ERROR could not read" + filename);
            else:
                exists = False

        return exists, timestamps

    def check_prereqs(self, required=True):
        return self.check_exist(self.requires, required)

    def check_postreqs(self, required=True):
        return self.check_exist(self.provides, required)

    def add_submodule(self, module):
        self.sub_modules.append(module)

    def inspect_internal_graph(self):
        self.D = generate_processing_graph(self.sub_modules)
        for item in self.D.find_sources():
            self.requires.add(item)
        for item in self.D.find_sinks():
            self.provides.add(item)



    def process_internal_graph(self):
        sorter = TopologicalSort(self.D)
        order = sorter.topological_sort()
        for item in order:
            if not isinstance(item, ProcessingModule):
                continue;
            item.process()




def boundaryCheck1(f):
    def checking_version(self):
        self.check_prereqs()
        f(self)
        self.check_postreqs()
    return checking_version

def boundaryCheck(f):
    def skipping_version(self, dry_run=False):
        if dry_run:
            pre_exist, pre_timestamps = self.check_prereqs(required=False)
            post_exist, post_timestamps = self.check_postreqs(required=False)
            if not pre_exist:
                print('[Dry Run] prereqs are not present for ' + self.name)
                return
            if post_exist and min(post_timestamps) > max(pre_timestamps):
                print("[Dry Run] Provide files up-to-date, will skip " + self.name)
                return;
            print('[Dry Run] Would need to run ' + self.name)
            return
            
        pre_exist, pre_timestamps = self.check_prereqs()
        post_exist, post_timestamps = self.check_postreqs(required=False)
        if post_exist and min(post_timestamps) > max(pre_timestamps):
            # can skip
            print("Provide files up-to-date, will skip " + self.name)
            return;

        print("Running " + self.name)
        f(self)
        self.check_postreqs()
    return skipping_version








def create_processing_module(func, name=None, files=None):
    """For specially definined functions, create a processing module class for them, as
    a convenience. Requirements:
      * The function must only take keyword arguments.
      * The keyword arguments refer to files or directories mostly
      * If it is a file the function creates, the files key starts with "out " 
      * If it is a file the function requires, the files key starts with "in ", and has a relative path
      * If it is a file the function requires, the files key starts with "in| ", and has an absolute path

    The function must also have a base and config keyword for those
    objects, but those do not need to be in the files dictionary

    The relative-path files are assumed to be relative to config['OUTPUT_DIR']
    Returns a classobj

    The relative/absolute path names may be templates, containing say
    {project}, which will be expanded using the contents of the config
    dictionary.

    """

    class Wrap(ProcessingModule):
        def __init__(self, base, config):
            ProcessingModule.__init__(self, base, config)
            self.name = func.__name__ if None == name else name
            self.files = {}
            self.func = func;

            func_args = set(inspect.getargs(func.__code__).args)
            def check_arg(argname):
                if argname not in func_args:
                        raise Exception("ERROR implementation function " + func.__name__ + 
                                        " does not have " + argname + " keyword parameter")


            if 'base' not in func_args:
                raise Exception("ERROR implementation function " + func.__name__ + " does not have base keyword parameter")
            if 'config' not in func_args:
                raise Exception("ERROR implementation function " + func.__name__ + " does not have config keyword parameter")

            for varname, pathname in files.items():
                pathname = pathname.format(**config)
                if varname.startswith("in "):
                    key = varname.split('in ')[1]
                    fullname = os.path.join(config['OUTPUT_DIR'], pathname)
                    self.requires.add(fullname)
                    self.files[key] = fullname
                    check_arg(key)
                    setattr(self, key, self.files[key])

                elif varname.startswith("in| "):
                    key = varname.split('in| ')[1]
                    self.requires.add(pathname)
                    self.files[key] = pathname
                    check_arg(key)
                    setattr(self, key, self.files[key])

                elif varname.startswith('out '):
                    key = varname.split('out ')[1]
                    fullname = os.path.join(config['OUTPUT_DIR'], pathname)
                    self.provides.add(fullname)
                    self.files[key] = fullname
                    check_arg(key)
                    setattr(self, key, self.files[key])

                elif 'base' == varname or 'config' == varname:
                    continue;
                else:
                    raise Exception("Unexpected files arg:" + varname)
                  

        @boundaryCheck
        def process(self):
            # automatically create output directories ???
            args = { k:i for k,i in self.files.items() }
            args['base'] = self.base
            args['config'] = self.config
            self.func(**args)
    return Wrap






def check_requirements(requirements):
    requirements_met = True;
    msg = "";
    for r in requirements:
        if isinstance(r, ProcessingModule):
            raise Exception("ERROR problem in graph where a processing module is a source requirement")
        if not os.access(r,  os.R_OK):
            msg += "ERROR not found " + r + "\n"
            requirements_met = False;
        else:
            print('Found ' + r)
    if not requirements_met:
        raise Exception(msg)


def make_it_go(L, output_graph=None, dry_run=False):
    D = generate_processing_graph(L)
    if None != output_graph:
        plot_graph(output_graph, D)

    # check
    requirements = D.find_sources() # find_initial_dependancies
    check_requirements(requirements)
    sorter = TopologicalSort(D)
    order = sorter.topological_sort()

    for item in order:
        if not isinstance(item, ProcessingModule):
            continue;
        item.process(dry_run=dry_run)



# Copyright 2016 Fred Hutchinson Cancer Research Center
# from distutils.core import setup

from setuptools import setup, find_packages


# or use find_packages()   from setuptools ??



packages = find_packages(exclude=['tests']);


setup(name='yagdspy',
      author='Keith Curtis',
      description='Graph dependancy system for processing tasks',
      license='License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
      version='1.0',
      url='https://github.com/krcurtis/yagdspy',
      classifiers=[
          'Development Status :: 4 - Beta',
          'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
          'Programming Language :: Python :: 3.5'
          ],
      keywords='graph dependancy',
      packages=packages,
      install_requires=['pydot'],
      python_requires='>=3.5.3',
      )
